/*
 * HANDS.C - generate 30 (or so) bridge hands in a compact format
 *
 * 21 June 1994 -- Released into the public domain.  
 * Authors Morris Jones and Matt Clegg.
 *
 * usage: hands [ -s <seed number> ]
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#ifdef MSDOS
#include <time.h>
#else
#include <sys/time.h>
#endif
#define _BRIDGE_
#include "types.h"

static char SCCSID[] = "@(#) hands.c 1.11@(#)";	/* SCCS keywords */

/* --------------------------------------------------------------------- */

/* Globals */

char print_buffer[13][98];	/* array holds one row of hands for printing */
long seed;
int seed_provided = 0;
deal cards[36];				/* the decks */
int starting_hand = 0;		/* starting hand for this page */
int porter_deals = 0;		/* if true, print dealing info */
char header[98];			/* page header */
char footer[98];			/* page footer */
char nl[] = "\n";
char ff[] = "\f";

/* --------------------------------------------------------------------- */

/* Forward references */

int main(int argc,char *argv[],char *envp[]);
void shuffle_the_deck(deal cards);
void srand48(long seed);
long lrand48(void);
void print_hand(int print_offset,deal cards,int hand_number);
void print_suit(int suit,int player,deal cards,int offset);
void print_page(int starting_hand);
void bprint(char *out,char *in);
void print_usage(void);
int getopt(int, char **, char *);
long atol(char *);
time_t time(time_t *);
void exit(int);
void print_porter(deal,int);

/* --------------------------------------------------------------------- */

/* Macros */

#ifdef MSDOS
#define RANDOM rand
#define SRANDOM(seed) srand(seed)
#else
#define RANDOM lrand48
#define SRANDOM(seed) srand48(seed)
#endif
#define random_mod(n) ((RANDOM()/64) % (n))

/* --------------------------------------------------------------------- */

int
main(int argc,char *argv[],char *envp[])
{
	int hand_number = 0;
	int i, line;
	int c;
	extern char *optarg;
	extern int optind;
	
/* Process the command line (not many options currently) */

	while ((c=getopt(argc,argv,"pqh:f:s:")) != -1) {
		switch (c) {
			case 's':
				seed_provided = 1;
				seed = atol(optarg);
				break;
			case 'p':
				porter_deals = 1;
				break;
			case 'q':
				porter_deals = 2;
				break;
			case 'h':
				memset(header,0,sizeof(header));
				strncpy(header,optarg,96);
				strcat(header,nl);
				break;
			case 'f':
				memset(footer,0,sizeof(footer));
				strncpy(footer,optarg,96);
				strcat(footer,nl);
				break;
			case '?':
			default:
				print_usage();
				exit(1);
				break;
		}
	}

/* shuffle 36 decks */

	if (!seed_provided) (void)time(&seed);
	SRANDOM(seed);

	for (i=0;i<36;++i) shuffle_the_deck(cards[i]);

/* Print the first page (15 hands) */

	hand_number = starting_hand;
	if (header[0]) {
		fputs(header,stdout);
		fputs(nl,stdout);
	}
	print_page(hand_number);
	if (footer[0]) {
		fputs(nl,stdout);
		fputs(footer,stdout);
	}
	fputs(ff,stdout);

/* Print the second page (15 hands) */

	hand_number += 15;
	if (header[0]) {
		fputs(header,stdout);
		fputs(nl,stdout);
	}
	print_page(hand_number);

/* Print the seed number at the bottom */

	if (footer[0]) {
		fputs(nl,stdout);
		fputs(footer,stdout);
	}
	printf("S = %lu\n",seed);

/* Print Porter deals if requested */

	if (porter_deals) {
		fputs(ff,stdout);
		fputs(header,stdout);
		fputs(nl,stdout);
		for (i=0;i<30;i++) print_porter(cards[i],i);
	}

	return 0;
}

/* --------------------------------------------------------------------- */

void
print_usage()
{
	printf("usage: hands [ options ]\n");
	printf("    -s seed   Use given number (long) for seed\n");
	printf("    -p        Print Gary Porter style dealing info after hands\n");
	printf("    -q        Print Gary Porter style, using ^>v<, not NeSw\n");
	printf("    -h \"hdr\"  Add the header following\n");
	printf("    -f \"ftr\"  Add the footer following\n");
}

/* --------------------------------------------------------------------- */

/* Prints five rows of three hands.  Includes some magic numbers that
 * relate to the column locations, and boundaries of vertical and
 * horizontal rules.  I don't feel the urge to generalize this beyond
 * this point right now.  print_hand() is more general, but still
 * refers to the global print buffer array.
 */

void 
print_page(int hand_number)
{
	int line, i;

/* Initialize the print buffer, including vertical rules */

	for (line =0; line<5; ++line) {
		for (i=0;i<13;++i) {
			memset(print_buffer[i],' ',96);
			print_buffer[i][96] = '\n';
			print_buffer[i][97] = 0;
		}
		for (i=0;i<12;++i) {
			print_buffer[i][30] = '|';
			print_buffer[i][62] = '|';
		}

/* Put three hands in the print buffer, replacing existing spaces */

		print_hand(0,cards[hand_number],hand_number);
		hand_number++;
		print_hand(32,cards[hand_number],hand_number);
		hand_number++;
		print_hand(64,cards[hand_number],hand_number);
		hand_number++;

/* Feed the print buffer to stdout, and print a separator line if
 * this isn't the last row of hands
 */

		if (line != 4) {
			memset(print_buffer[12],'-',92);
			print_buffer[12][30] = '+';
			print_buffer[12][62] = '+';
			for (i=0;i<13;++i) {
				fputs(print_buffer[i],stdout);
			}
		}
		else {
			for (i=0;i<12;++i) {
				fputs(print_buffer[i],stdout);
			}
		}
	}
}

/* --------------------------------------------------------------------- */

/* print one hand at specified x offset
 *
 * This routine takes full advantage of the data structures designed
 * by Matt Clegg for bridge hands, suits, players, and vulnerabilities.
 * It references the print buffer as an x,y character positions, adding
 * the x offset specified.  (Future expansion might add a y offset as
 * well.)
 */

void 
print_hand(int print_offset,deal cards,int hand_number)
{
	int player, suit;
	int i;
	int x,y;
	char *p;
	char *dlr;
	char *vul;
	int vul_index;
	char buf[20];

/* Create strings for dealer and vulnerability */

	switch (dealer_list[hand_number % 16]) {
		case PLAYER_NORTH:
			dlr = "N";
			break;
		case PLAYER_EAST:
			dlr = "E";
			break;
		case PLAYER_SOUTH:
			dlr = "S";
			break;
		case PLAYER_WEST:
			dlr = "W";
			break;
	}

	vul_index = ns_vulnerability_list[hand_number % 16] + 
	  2 * ew_vulnerability_list[hand_number % 16];

	switch (vul_index) {
		case 0:
			vul = "None";
			break;
		case 1:
			vul = "N-S";
			break;
		case 2:
			vul = "E-W";
			break;
		case 3:
			vul = "Both";
			break;
	}

/* Install board number, dealer, and vulnerability in the print buffer */

	p = &print_buffer[0][0+print_offset];
	sprintf(buf,"Bd  %d",hand_number + 1);
	bprint(p,buf);

	p = &print_buffer[1][0+print_offset];
	sprintf(buf,"Dlr %s",dlr);
	bprint(p,buf);

	p = &print_buffer[2][0+print_offset];
	sprintf(buf,"Vul %s",vul);
	bprint(p,buf);

/* Loop on the players and the suits to install the cards themselves */

	for ( player = 0; player < 4 ; ++player) {
		for ( suit = 0; suit < 4; ++suit) {
			print_suit(suit,player,cards,print_offset);
		}
	}
}

/* --------------------------------------------------------------------- */

/* print one suit for one hand
 *
 * This routine references an array of x,y coordinates that is indexed
 * by player and suit.  suit_offsets[][] is defined in types.h
 */

void 
print_suit(int suit,int player,deal cards,int offset)
{
	int i;
	char *p;
	int x, y;
	char buf[10];

	x = suit_offsets[player][suit].x + offset;
	y = suit_offsets[player][suit].y;
	p = &print_buffer[y][x];

	*p++ = *suit_names[suit];
	p++;
	for ( i=12; i >= 0; --i ) {
		if (cards[i + 13*suit] == player) {
			*p++ = *rank_names[i];
		}
	}
}

/* --------------------------------------------------------------------- */

/* Use this routine instead of sprintf() to avoid having a NULL term */

void 
bprint(char *out,char *in)
{
	while (*in)  *out++ = *in++;
}

/* --------------------------------------------------------------------- */

/* Print hands in Gary Porter style:
 * Board 1  NeWssWNNW
 */

void
print_porter(deal cards,int board)
{
	int card;
	int suit;
	int marker[4];	/* four player markers */

	if (porter_deals == 1) {
		marker[PLAYER_NORTH] = 'N';
		marker[PLAYER_EAST] = 'e';
		marker[PLAYER_SOUTH] = 'S';
		marker[PLAYER_WEST] = 'w';
	}
	else if (porter_deals == 2) {
		marker[PLAYER_NORTH] = '^';
		marker[PLAYER_EAST] = '>';
		marker[PLAYER_SOUTH] = 'v';
		marker[PLAYER_WEST] = '<';
	}

	printf("Board %2d   ",board+1);

	for (suit = SUIT_SPADES; suit >= 0; --suit) {
		for (card = 12; card >= 0; --card) {
			switch (cards[card+13*suit]) {
				case PLAYER_NORTH:
					putchar(marker[PLAYER_NORTH]);
					break;
				case PLAYER_EAST:
					putchar(marker[PLAYER_EAST]);
					break;
				case PLAYER_SOUTH:
					putchar(marker[PLAYER_SOUTH]);
					break;
				case PLAYER_WEST:
					putchar(marker[PLAYER_WEST]);
					break;
			}
		}
		putchar(' ');
	}
	printf("\n\n\n");
}

/* --------------------------------------------------------------------- */

/* Almost direct from okbridge -- the seed is installed up in main()
 */

void 
shuffle_the_deck (deal cards)
/* Using the algorithm suggested by Douglas Foxvog.  Thanks, Doug! */
{
	int i, t, c;
	deal shuffle;

	for (i = 0; i < 52; i++) 
		shuffle [i] = i;
	for (i = 0; i < 51; i++) {
		c = random_mod (52 - i);
		t = shuffle[i+c]; 
		shuffle[i+c] = shuffle[i];
		shuffle[i] = t;
	};
	for (i = 0; i < 52; i++)
		cards[shuffle[i]] = (i % 4);
		
}
